package ostate;

import ostate.util.tree.StatesTree;
import ostate.util.tree.Node;
import ostate.util.properties.Properties;
import ostate.invoke.*;
import ostate.invoke.impl.DefaultEventProcessor;
import ostate.invoke.impl.DefaultMethodInvoker;
import ostate.exceptions.InvalidMethodCallException;
import ostate.exceptions.InternalErrorException;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.*;

/**
 * Invocation handler, used by State Machine proxy;
 *
 * @author Pavel Mavrin
 * @version $Id$
 */
class StateMachineManager<I> implements InvocationHandler {

    public static Method STATE_START_METHOD;

    public static Method STATE_STOP_METHOD;

    {
        for (Method method : IState.class.getMethods()) {
            if (method.getName().equals("stateEntry")) {
                STATE_START_METHOD = method;
            }
            if (method.getName().equals("stateExit")) {
                STATE_STOP_METHOD = method;
            }
        }
        assert STATE_START_METHOD != null;
        assert STATE_STOP_METHOD != null;
    }

    private final StatesTree<I> statesTree;

    private final StatesTreeController<I> statesTreeController;

    private final StateMachineImpl stateMachine = new StateMachineImpl();

    private MethodInvoker defautMethodInvoker = DefaultMethodInvoker.CHILDREN_ROOT;

    private EventProcessor defautEventProcessor = DefaultEventProcessor.CHILDREN_ROOT;

    public StateMachineManager(Class<I> AI, IStateFactory<I> sf) {
        statesTree = new StatesTree<I>();
        statesTreeController = new StatesTreeController<I>(statesTree, sf, AI, stateMachine.sources);
    }

    public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
        try {
            IStateMachine.class.getMethod(method.getName(), method.getParameterTypes());
        } catch(NoSuchMethodException e) {
            try {
                return stateMachine.invokeMethod(method, args);
            } catch(IllegalAccessException ex) {
                throw new InternalErrorException("Cannot access method.", ex);
            } catch(InvocationTargetException ex) {
                throw ex.getTargetException();
            }
        }
        try {
            return method.invoke(stateMachine, args);
        } catch(InvocationTargetException e) {
            throw e.getTargetException();
        }
    }


    private class StateMachineImpl implements IStateMachine<I> {

        final WorkQueues workQueues = new WorkQueues();

        Stack<IState<I>> sources = new Stack<IState<I>>();

        public void changeState(Class<? extends IState<I>> target, Properties properties, Runnable action) {
            boolean my = false;
            synchronized(workQueues) {
                if (workQueues.inUse) {
                    if (OStateProperties.isDebug())
                        System.out.println("tr: " + sources.peek() + " -> " + target);
                    workQueues.transitionQueue.offer(new Transition<I>(
                            sources.peek(),
                            target,
                            properties,
                            action
                    ));
                } else {
                    if (OStateProperties.isDebug())
                        System.out.println("tr: -> " + target);
                    workQueues.inUse = true;
                    my = true;
                    workQueues.transitionQueue.offer(new Transition<I>(
                            null,
                            target,
                            properties,
                            action
                    ));
                }
            }
            if (my) {
                processAll();
            }
        }

        public void changeState(Class<? extends IState<I>> newStateClass, Properties properties) {
            changeState(newStateClass, properties, null);
        }

        public void changeState(Class<? extends IState<I>> newStateClass) {
            changeState(newStateClass, null, null);
        }

        public void processEvent(Event event, EventProcessor eventProcessor) {
            boolean my = false;
            synchronized(workQueues) {
                workQueues.eventQueue.offer(new EventProcessingPair(
                        event,
                        eventProcessor)
                );
                if (!workQueues.inUse) {
                    workQueues.inUse = true;
                    my = true;
                }
            }
            if (my) {
                processAll();
            }
        }

        private void processAll() {
            while (true) {
                while (!workQueues.eventQueue.isEmpty() || !workQueues.transitionQueue.isEmpty()) {
                    if (!workQueues.eventQueue.isEmpty()) {
                        EventProcessingPair p = workQueues.eventQueue.remove();
                        Event event = p.event;
                        EventProcessor processor = p.processor;
                        statesTreeController.processEvent(
                                statesTree.getRoot(),
                                event,
                                processor);
                    }
                    if (OStateProperties.isDebug())
                        System.out.println("EQE");
                    while (!workQueues.transitionQueue.isEmpty()) {
                        Transition<I> transition = workQueues.transitionQueue.remove();
                        statesTreeController.processTransition(transition);
                    }
                    if (OStateProperties.isDebug())
                        System.out.println("TQE");
                }
                if (OStateProperties.isDebug())
                    System.out.println("AQE");
                synchronized (workQueues) {
                    if (workQueues.eventQueue.isEmpty() &&
                            workQueues.transitionQueue.isEmpty()) {
                        workQueues.inUse = false;
                        break;
                    }
                }
            }
        }

        public void processEvent(Event event) {
            processEvent(event, defautEventProcessor);
        }

        public <T extends IState<I>> T getCurrentStateByClass(Class<T> stateClass) {
            Node<IState<I>> node = statesTree.getNode(stateClass);
            if (node == null) {
                return null;
            } else {
                return (T)node.getValue();
            }
        }

        public Object invokeMethod(Method method, Object[] args)
                throws IllegalAccessException, InvocationTargetException {
            if (!workQueues.inUse) {
                throw new InvalidMethodCallException(
                        "Cannot call method from inside"
                );
            } else {
                workQueues.inUse = true;
                Object result = statesTreeController.invokeMethod(
                        statesTree.getRoot(),
                        method,
                        defautMethodInvoker,
                        null, args);
                processAll();
                workQueues.inUse = false;
                return result;
            }
        }

        private class EventProcessingPair {
            final Event event;
            final EventProcessor processor;

            public EventProcessingPair(Event event, EventProcessor processor) {
                this.event = event;
                this.processor = processor;
            }
        }

        private class WorkQueues {

            boolean inUse = false;

            Queue<EventProcessingPair> eventQueue = new LinkedList<EventProcessingPair>();

            Queue<Transition<I>> transitionQueue = new LinkedList<Transition<I>>();

        }

    }

}