package ostate.discriptions;

import ostate.Event;
import ostate.IState;
import ostate.exceptions.InvalidStateException;
import ostate.annotations.Handler;
import ostate.annotations.State;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.Map;
import java.util.Collections;

/**
 * Factory, that produces state discriptors, using their annotations.
 *
 * Factory uses {@link State} annotation to recognize
 * a type of a state, and {@link Handler} annotation
 * to recognize the handlers.
 *
 * @author Pavel Mavrin
 * @version $Id$
 */
public class StateDiscriptorFactory {

    private final static Map<Class<? extends IState>, StateDiscriptor> known =
            new HashMap<Class<? extends IState>, StateDiscriptor>();

    public static <I> StateDiscriptor<I> getStateDiscriptor(IState<I> istate) {
        return getStateDiscriptor((Class<? extends IState<I>>)istate.getClass());
    }

    public static <I> StateDiscriptor<I> getStateDiscriptor(Class<? extends IState<I>> stateClass) {
        if (known.keySet().contains(stateClass)) {
            return known.get(stateClass);
        }
        if (!stateClass.isAnnotationPresent(State.class)) {
            throw new InvalidStateException("IState is not annotated", stateClass);
        }
        State state = stateClass.getAnnotation(State.class);
        StateDiscriptor<I> stateDiscriptor = null;
        switch (state.type()) {
            case SIMPLE:
                stateDiscriptor = new SimpleStateDiscriptor<I>(state);
                break;
            case ANDSTATE:
                stateDiscriptor = new ANDStateDiscriptor<I>(state);
                break;
            case ORSTATE:
                stateDiscriptor = new ORStateDiscriptor<I>(state);
                break;
        };
        addHandlers(stateDiscriptor, stateClass);
        Collections.sort(stateDiscriptor.getHandlers());
        known.put(stateClass, stateDiscriptor);
        return stateDiscriptor;
    }

    private static <I> void addHandlers(StateDiscriptor<I> discriptor, Class stateClass) {
        for (Method method : stateClass.getMethods()) {
            if (method.isAnnotationPresent(Handler.class)) {
                Class[] parameters = method.getParameterTypes();
                if (parameters.length != 1) {
                    throw new InvalidStateException("Handler " + method.toString() + " have too much parameters", stateClass);
                }
                if (!Event.class.isAssignableFrom(parameters[0])) {
                    throw new InvalidStateException("Handler " + method.toString() + " paremeter is not a subclass of ostate.Event", stateClass);
                }
                if (method.getExceptionTypes().length > 0) {
                    throw new InvalidStateException("Handler " + method.toString() + " throws exception", stateClass);
                }
                if ((method.getModifiers() & Modifier.PUBLIC) == 0) {
                    throw new InvalidStateException("Handler " + method.toString() + " is not public", stateClass);
                }
                discriptor.addHandler(parameters[0], method);
            }
        }
    }
}
